/*
 *	This file implements a netcf interface
 *
 *	Copyright (C) 2010 Olaf Kirch <okir@suse.de>
 *	Copyright (C) 2011-2013 SUSE LINUX Products GmbH, Nuernberg, Germany.
 *
 *	This library is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU Lesser General Public
 *	License as published by the Free Software Foundation; either
 *	version 2.1 of the License, or (at your option) any later version.
 *
 *	This library is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with this library; if not, write to the Free Software
 *	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *	MA  02110-1301  USA
 *
 *	Authors: Olaf Kirch <okir@suse.de>
 *	         Marius Tomaschewski <mt@suse.de>
 *
 *	libnetcontrol contains source code which is based on wicked.
 *	Wicked is licensed under the GPL-2.0+, but permission has been
 *	granted by the authors of wicked to use the code derived from
 *	wicked under the LGPL-2.1+ in libnetcontrol.
 *	You can find the wicked project at http://gitorious.org/wicked/.
 *
 */
#if defined(HAVE_CONFIG_H)
#include <config.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>

#include <netcontrol/netcf.h>
#include <sutils.h>
#include <nutils.h>
#include <handle.h>
#include <syntax.h>
#include <logging.h>
#include <unistd.h>

#define __NETCONTROL_NETCF_DEBUG	1

/* === global data ======= */
// mutex
// write lock

/* === data structures === */
struct netcf_ifinfo {
	nc_handle_t *		handle;
};

struct netcf {
	unsigned int		users;
	int			error;
	char *			errstr;

	struct netcf_ifinfo	system;
	struct netcf_ifinfo 	config;
	nc_syntax_t *		syntax;
};

struct netcf_if {
	struct netcf *		netcf;
	nc_interface_t *	ifp;
	nc_stringbuf_t		mac;
};


/* === private prototypes === */
static const char *		__ncf_strerror(int error);
static int			__ncf_list_interfaces(struct netcf *ncf, unsigned int flags,
				                      char **names, unsigned int nnames);
static struct netcf_if *	__ncf_if_new(struct netcf *ncf, nc_interface_t *ifp);
static nc_interface_t *		__ncf_if_state(struct netcf_if *nif);

static char *			__ncf_if_xml(struct netcf *ncf, struct netcf_ifinfo *nii,
						nc_interface_t *ifp);
static int			__ncf_refresh(struct netcf_ifinfo *nii);

static void			__ncf_error_clear (struct netcf *ncf, struct netcf_ifinfo *nii);
static void			__ncf_error_report(struct netcf *ncf, struct netcf_ifinfo *nii,
							int error, const char *fmt, ...);

/* === public functions === */
int
ncf_init(struct netcf **netcf, const char *root)
{
	struct netcf *ncf = NULL;

	if(!netcf)
		return -1;
	*netcf = NULL;

	if(getuid() != 0) {
		/*
		** - All ifcfg files are readable by root only (passwd)
		** - /proc/vlan/config is readable by root only
		** - Every modifying action need root permissions
		*/
		nc_error("UID root permissions needed to read config and some states");
		return -1;
	}

	ncf = calloc(1, sizeof(*ncf));
	if( !ncf)
		goto failure;
	ncf->users = 1;

	ncf->system.handle = nc_indirect_open("/system", root);
	if(!ncf->system.handle)
		goto failure;

	ncf->config.handle = nc_indirect_open("/config", root);
	if(!ncf->config.handle)
		goto failure;

	ncf->syntax = nc_syntax_new("netcf", NULL);
	if(!ncf->syntax)
		goto failure;

	/*
	** we ignore NETCF_DATADIR & NETCF_DEBUG env vars.
	*/

	*netcf = ncf;
	return 0;

failure:
	ncf_close(ncf);;
	return -1;
}


int
ncf_close(struct netcf *ncf)
{
	if(ncf) {
		if(ncf->users != 1) {
			ncf->error = NETCF_EINUSE;
			return -1;
		}
		if(ncf->system.handle)
			nc_close(ncf->system.handle);
		if(ncf->config.handle)
			nc_close(ncf->config.handle);
		if(ncf->syntax)
			nc_syntax_free(ncf->syntax);

		nc_string_free(&ncf->errstr);

		memset(ncf, 0, sizeof(*ncf));
		free(ncf);
	}
	return 0;
}


int
ncf_error(struct netcf *ncf, const char **errmsg, const char **details)
{
	int error = ncf->error;
	if (errmsg)
		*errmsg = __ncf_strerror(error);
	if (details)
		*details = ncf->errstr;
	ncf->error = 0;
	return error;
}


int
ncf_num_of_interfaces(struct netcf *ncf, unsigned int flags)
{
	return __ncf_list_interfaces(ncf, flags, NULL, 0);
}


int
ncf_list_interfaces(struct netcf *ncf, int maxnames, char **names, unsigned int flags)
{
	if(names && maxnames > 0) {
		memset(names, 0, maxnames * sizeof(*names));
	} else {
		maxnames = 0;
		names = NULL;
	}
	return __ncf_list_interfaces(ncf, flags, names, maxnames);
}


struct netcf_if *
ncf_lookup_by_name(struct netcf *ncf, const char *name)
{
	nc_interface_t *ifp, *sfp;
	struct netcf_if *nif;
	int refreshed = 0;

	if (__ncf_refresh(&ncf->config) < 0) {
		__ncf_error_report(ncf, &ncf->config, NETCF_EOTHER,
			"Cannot read stored configuration");
		return NULL;
	}

	ifp = nc_interface_by_name(ncf->config.handle, name);
	if(ifp && !ifp->invalid) {

		if(nc_syntax_supported_inteface(ncf->syntax, ncf->config.handle, ifp) != 0) {
			nc_info("skipping syntax unsupported interface %s", ifp->name);
		} else {
			/* OK, we have a supported interface */
			nif = __ncf_if_new(ncf, ifp);
			if(nif && !nc_hw_address_format(&ifp->hwaddr, &nif->mac)) {
				if(!refreshed) {
					refreshed = 1;
					if (__ncf_refresh(&ncf->system) < 0)
						return nif;
				}
				sfp = nc_interface_find(ncf->system.handle, NULL, -1, ifp->name, 0, 0);
				if(sfp)
					nc_hw_address_format(&sfp->hwaddr, &nif->mac);
			}
			return nif;
		}

	}
	/* no error, just there is no such interface */
	return NULL;
}


int
ncf_lookup_by_mac_string(struct netcf *ncf, const char *mac, int maxifaces,
			 struct netcf_if **ifaces)
{
	nc_interface_t *ifp;
	unsigned int pos;
	int count = 0;

	if(!ncf || !mac) {
		ncf->error = NETCF_EOTHER;
		return -1;
	}

	if(ifaces && maxifaces > 0) {
		memset(ifaces, 0, maxifaces * sizeof(*ifaces));
	} else {
		ifaces = NULL;
		maxifaces = 0;
	}

	if (__ncf_refresh(&ncf->config) < 0) {
		__ncf_error_report(ncf, &ncf->config, NETCF_EOTHER,
			"Cannot read stored configuration");
		return -1;
	}

	for(ifp = nc_interface_first(ncf->config.handle, &pos); ifp;
	    ifp = nc_interface_next(ncf->config.handle, &pos))
	{
		nc_stringbuf_t nsb = NC_STRINGBUF_INIT;

		if(ifp->invalid)
			continue;

		if(nc_syntax_supported_inteface(ncf->syntax, ncf->config.handle, ifp) != 0) {
			nc_info("skipping syntax unsupported interface %s", ifp->name);
			continue;
		}

		if(nc_hw_address_format(&ifp->hwaddr, &nsb) == NULL) {
			nc_interface_t *sfp = nc_interface_find(ncf->system.handle, NULL, -1, ifp->name, 0, 0);
			if(sfp) {
				if(nc_hw_address_format(&sfp->hwaddr, &nsb) == NULL)
					continue;
			}
		}

		if(!nc_string_eq(mac, nsb.string)) {
			nc_stringbuf_destroy(&nsb);
			continue;
		}

		if(count < maxifaces) {
			ifaces[count] = __ncf_if_new(ncf, ifp);
			if(ifaces[count]) {
				nc_stringbuf_move(&ifaces[count]->mac, &nsb);
			} else {
				nc_stringbuf_destroy(&nsb);
				break;
			}
		} else {
			nc_stringbuf_destroy(&nsb);
		}
		count++;
	}
	return count;
}

#ifdef __NETCONTROL_NETCF_DEBUG
static inline void __ncf_parsed_xml(const char *str, void *arg) 
{
	(void)arg;
	nc_debug_netcf("xml: %s", str);
}
#endif

struct netcf_if *
ncf_define(struct netcf *ncf, const char *xml_data)
{
	FILE           *memstream;
	nc_interface_t *cfg, *ifp;
	xml_node_t     *xml_root;
	xml_node_t     *ifnode;
	struct netcf_if *nif;

#ifdef __NETCONTROL_NETCF_DEBUG
	nc_debug_netcf("define xml=\n%s", xml_data);
#endif
	memstream = fmemopen((char *) xml_data, strlen(xml_data), "r");
	if(memstream == NULL) {
		__ncf_error_report(ncf, NULL, NETCF_EOTHER,
			"Unable to open memstream for xml data: %m");
		return NULL;
	}

	xml_root = xml_node_scan(memstream);
	fclose(memstream);
	if (xml_root == NULL) {
		__ncf_error_report(ncf, NULL, NETCF_EXMLPARSER,
			"Unable to parse xml data: %s", xml_data);
		return NULL;
	}

	if(nc_string_eq(xml_root->name, "interface")) {
		ifnode = xml_root;
	} else {
		ifnode = xml_node_get_child(xml_root, "interface");
	}
	if(ifnode == NULL) {
		__ncf_error_report(ncf, NULL, NETCF_EXMLINVALID,
			"Unable to find interface node in xml data: %s",
		       	xml_data);
		xml_node_free(xml_root);
		return NULL;
	}

#ifdef __NETCONTROL_NETCF_DEBUG
	xml_node_print_fn(ifnode, __ncf_parsed_xml, NULL);
#endif

	__ncf_error_clear(ncf, &ncf->config);
	if (__ncf_refresh(&ncf->config) < 0) {
		__ncf_error_report(ncf, &ncf->config, NETCF_EOTHER,
			"Cannot reread stored configuration");
		xml_node_free(xml_root);
		return NULL;
	}

	cfg = nc_syntax_xml_to_interface(ncf->syntax, ncf->config.handle, ifnode);
	xml_node_free(xml_root);
	if(cfg == NULL) {
		__ncf_error_report(ncf, &ncf->config, NETCF_EXMLINVALID,
			"Cannot to interpret xml interface configuration: %s", xml_data);
		return NULL;
	}

	if(nc_interfaces_configure(ncf->config.handle, cfg) < 0) {
		__ncf_error_report(ncf, &ncf->config, NETCF_EOTHER,
			"Configuration of interface '%s' failed", cfg->name);

		__ncf_refresh(&ncf->config);
		nc_interface_free(cfg);
		return NULL;
	}

#if 1	/* reread all config */
	if (__ncf_refresh(&ncf->config) < 0) {
		__ncf_error_report(ncf, &ncf->config, NETCF_EOTHER,
			"Cannot reread stored configuration");
		return NULL;
	}
#endif

	ifp = nc_interface_by_name(ncf->config.handle, cfg->name);
	if(ifp == NULL) {
		__ncf_error_report(ncf, &ncf->config, NETCF_EOTHER,
			"Configuration of interface '%s' failed", cfg->name);
		nc_interface_free(cfg);
		return NULL;
	}
	nc_interface_free(cfg);

#if 1	/* regression test */
	ifnode = nc_syntax_xml_from_interface(ncf->syntax, ncf->config.handle, ifp);
	if(ifnode == NULL) {
		__ncf_error_report(ncf, &ncf->config, NETCF_EOTHER,
			"Configuration of interface '%s' failed", cfg->name);
		return NULL;
	}
	xml_node_free(ifnode);
#endif

	nif = __ncf_if_new(ncf, ifp);
	if(!nif) {
		__ncf_error_report(ncf, &ncf->config, NETCF_EOTHER,
			"Configuration of interface '%s' failed", cfg->name);
		return NULL;
	}
	return nif;
}

void
ncf_if_free(struct netcf_if *nif)
{
	if(!nif)
		return;
	assert(nif->netcf && nif->netcf->users > 1);

	if (nif->ifp)
		nc_interface_free(nif->ifp);
	nc_stringbuf_destroy(&nif->mac);

	nif->netcf->users--;
	memset(nif, 0, sizeof(*nif));
	free(nif);
}

const char *
ncf_if_name(struct netcf_if *nif)
{
	nc_interface_t *ifp;

	if (!(ifp = nif->ifp))
		return NULL;

	if (!ifp->name)
		return NULL;

	return ifp->name;
}


const char *
ncf_if_mac_string(struct netcf_if *nif)
{
	if(nif) {
		if(nif->mac.len)
			return nif->mac.string;

		if(nif->ifp)
			return nc_hw_address_format(&nif->ifp->hwaddr, &nif->mac);
	}
	return NULL;
}


char *
ncf_if_xml_desc(struct netcf_if *nif)
{
	if(!nif || !nif->netcf || !nif->ifp)
		return NULL;

	char *xml = __ncf_if_xml(nif->netcf, &nif->netcf->config, nif->ifp);
#ifdef __NETCONTROL_NETCF_DEBUG
	nc_debug_netcf("descr xml for %s:\n%s", nif->ifp->name, xml);
#endif
	return xml;
}

char *
ncf_if_xml_state(struct netcf_if *nif)
{
	nc_interface_t *ifp;

	if(!nif || !nif->netcf || !nif->ifp)
		return NULL;

	if ((ifp = __ncf_if_state(nif)) == NULL) {
		/* consider it down */
		return NULL;
	}

	char *xml = __ncf_if_xml(nif->netcf, &nif->netcf->system, ifp);
#ifdef __NETCONTROL_NETCF_DEBUG
	nc_debug_netcf("state xml for %s:\n%s", nif->ifp->name, xml);
#endif
	return xml;
}


int
ncf_if_up(struct netcf_if *nif)
{
	assert(nif && nif->netcf && nif->ifp && nif->ifp->name);

	/* Note, we use config handle here (iface may not exist yet) */
	if(nc_interface_start(nif->netcf->config.handle, nif->ifp) == 0) {
		nif->netcf->error = 0;
		return 0;
	} else {
		__ncf_error_report(nif->netcf, &nif->netcf->config, NETCF_EOTHER,
			"Unable to start interface '%s'", nif->ifp->name);
		return -1;
	}
}

int
ncf_if_down(struct netcf_if *nif)
{
	assert(nif && nif->netcf && nif->ifp && nif->ifp->name);

	/* Note, we use config handle here (no need to refresh state) */
	if(nc_interface_stop(nif->netcf->config.handle, nif->ifp) == 0) {
		nif->netcf->error = 0;
		return 0;
	} else {
		__ncf_error_report(nif->netcf, &nif->netcf->config, NETCF_EOTHER,
			"Unable to stop interface '%s'", nif->ifp->name);
		return -1;
	}
}

int
ncf_if_status(struct netcf_if *nif, unsigned int *flags)
{
	nc_interface_t *ifp;

	assert(nif && nif->netcf);
	if(!nif->ifp || !nif->ifp->name) {
		__ncf_error_report(nif->netcf, NULL, NETCF_EOTHER,
			"Invalid interface handle");
		return -1;
	} else if(!flags) {
		__ncf_error_report(nif->netcf, NULL, NETCF_EOTHER,
			"NULL flags pointer in ncf_if_status");
		return -1;
	}

	/* Hmm... We don't call ifstatus -- do we have to? */
	*flags = NETCF_IFACE_INACTIVE;
	if (!(ifp = __ncf_if_state(nif))) {
		/* No error, just consider it down */
		nif->netcf->error = 0;
#if 0
	} else if (nc_interface_network_is_up(ifp)) {
		*flags = NETCF_IFACE_ACTIVE;
		nif->netcf->error = 0;
#endif
	} else {
		int ret = nc_interface_status(nif->netcf->system.handle, nif->ifp);
		if(ret < 0) {
			__ncf_error_report(nif->netcf, &nif->netcf->system, NETCF_EOTHER,
				"Cannot get status of interface %s",
				nif->ifp->name);
			return -1;
		}
		if(ret == 0) {
			*flags = NETCF_IFACE_ACTIVE;
			nif->netcf->error = 0;
		}
	}
	return 0;
}

int
ncf_if_undefine(struct netcf_if *nif)
{
	if(nif && nif->netcf) {
		if(nif->ifp && nif->ifp->name) {
			if(nc_interfaces_delete(nif->netcf->config.handle, nif->ifp) == 0) {
				nif->netcf->error = 0;
				return 0;
			}
			__ncf_error_report(nif->netcf, &nif->netcf->config, NETCF_EOTHER,
				"Unable to remove configuration for interface '%s'",
					nif->ifp->name);
		} else {
			__ncf_error_report(nif->netcf, &nif->netcf->config, NETCF_EOTHER,
				"Invalid interface reference");
		}
	}
	return -1;
}


/* === private helpers === */
static const char *
__ncf_strerror(int error)
{
	switch (error) {
	case NETCF_NOERROR:
		return "no error";
	case NETCF_EINTERNAL:
		return "internal error";
	case NETCF_EOTHER:
		return "other error";
	case NETCF_ENOMEM:
		return "allocation failed";
	case NETCF_EXMLPARSER:
		return "XML parser choked";
	case NETCF_EXMLINVALID:
		return "XML invalid in some form";
	case NETCF_ENOENT:
		return "Required entry in a tree is missing";
	case NETCF_EEXEC:
		return "external program execution failed";
	case NETCF_EINUSE:
		return "cannot close netcf, instances still in use";
	case NETCF_EXSLTFAILED:
		return "XSLT transformation failed";
	case NETCF_EFILE:
		return "some file access failed";
	case NETCF_EIOCTL:
		return "an ioctl call failed";
	case NETCF_ENETLINK:
		return "netlink error";
	default:
		return "error code not handled";
	}
}

static void
__ncf_error_clear(struct netcf *ncf, struct netcf_ifinfo *nii)
{
	if(ncf) {
		ncf->error = 0;
		nc_string_free(&ncf->errstr);
		if(nii) {
			nc_error_detail_clear(nii->handle);
		}
	}
}

static void
__ncf_error_report(struct netcf *ncf, struct netcf_ifinfo *nii,
			int error, const char *fmt, ...)
{
	if(ncf) {
		/* Hmm... no details for success */
		nc_string_free(&ncf->errstr);
		if((ncf->error = error) == NETCF_NOERROR)
			return;

		/* When handle provided and has details, use them */
		if(nii && nii->handle) {
			nc_error_detail_move(nii->handle, &ncf->errstr);
			if(ncf->errstr)
				return;
		}

		/* Otherwise, format the provided message */
		if(fmt) {
			nc_stringbuf_t  buf = NC_STRINGBUF_INIT;
			va_list         ap;     

			va_start(ap, fmt);
			nc_stringbuf_vprintf(&buf, fmt, ap);
			va_end(ap);

			ncf->errstr = buf.string;
		}
	}
}

static int
__ncf_list_interfaces(struct netcf *ncf, unsigned int flags,
                        char **names, unsigned int nnames)
{
	static const unsigned int match_any = (NETCF_IFACE_ACTIVE|NETCF_IFACE_INACTIVE);
	nc_interface_t *ifp;
	unsigned int count = 0, pos;
	nc_handle_t *nh;

	if (__ncf_refresh(&ncf->config) < 0) {
		__ncf_error_report(ncf, &ncf->config, NETCF_EOTHER,
			"Cannot read stored configuration");
		return -1;
	}

	/* If we're asking for either active or inactive,
	 * also need to refresh the state
	 */
	if( (flags & match_any) != match_any) {
		nh = ncf->config.handle;

		if (__ncf_refresh(&ncf->system) < 0) {
			__ncf_error_report(ncf, &ncf->system, NETCF_EOTHER,
				"Cannot read system interface state");
			return -1;
		}

		for(ifp = nc_interface_first(nh, &pos); ifp;
		    ifp = nc_interface_next(nh, &pos)) {
			nc_interface_t *cur;
			unsigned int state;

			if(ifp->invalid)
				continue;

			if(nc_syntax_supported_inteface(ncf->syntax, nh, ifp) != 0) {
				nc_info("skipping syntax unsupported interface %s",
					ifp->name);
				continue;
			}
			if(ifp->parent) {
				nc_info("skipping iface %s with parent %s",
					ifp->name, ifp->parent);
				continue;
			}

			state = NETCF_IFACE_INACTIVE;
			cur = nc_interface_by_name(ncf->system.handle, ifp->name);
			if(cur) {
				nc_interface_status(ncf->system.handle, cur);
				if(nc_interface_network_is_up(cur))
					state = NETCF_IFACE_ACTIVE;
			}

			if(flags & state) {
				if(count < nnames) {
					if(nc_string_dup(&names[count], ifp->name) != 0)
						break;
				}
				count++;
			}
		}
	} else {
		nh = ncf->config.handle;

		for(ifp = nc_interface_first(nh, &pos); ifp;
		    ifp = nc_interface_next(nh, &pos))	{

			if(ifp->invalid)
				continue;

			if(nc_syntax_supported_inteface(ncf->syntax, nh, ifp) != 0) {
				nc_info("skipping syntax unsupported interface %s",
					ifp->name);
				continue;
			}
			if(ifp->parent) {
				nc_info("skipping iface %s with parent %s",
					ifp->name, ifp->parent);
				continue;
			}

			if (count < nnames) {
				if(nc_string_dup(&names[count], ifp->name) != 0)
					break;
			}
			count++;
		}
	}

	return count;
}


static struct netcf_if *
__ncf_if_new(struct netcf *ncf, nc_interface_t *ifp)
{
	struct netcf_if *nif;

	if(!ncf || !ifp)
		return NULL;

	nif = calloc(1, sizeof(*nif));
	if(nif) {
		nif->netcf = ncf;
		nif->ifp   = nc_interface_ref(ifp);
		nc_stringbuf_init(&nif->mac);
		ncf->users++;
	}
	return nif;
}

static char *
__ncf_if_xml(struct netcf *ncf, struct netcf_ifinfo *nii, nc_interface_t *ifp)
{
	FILE   *mem_stream = NULL;
	char   *out_string = NULL;
	size_t  out_size = 0;
	xml_node_t *xml = NULL;

	xml = nc_syntax_xml_from_interface(ncf->syntax, nii->handle, ifp);
	if(!xml) {
		__ncf_error_report(ncf, nii, NETCF_EOTHER,
			"Failed to generate xml representation of interface '%s'",
			(ifp && ifp->name ? ifp->name : NULL));
		goto error;
	}

	mem_stream = open_memstream(&out_string, &out_size);
	if(!mem_stream) {
		__ncf_error_report(ncf, NULL, NETCF_EOTHER,
			"Failed to open memory stream for xml representation of interface '%s'",
			(ifp && ifp->name ? ifp->name : NULL));
		goto error;
	}

	if (xml_node_print(xml, mem_stream) < 0) {
		__ncf_error_report(ncf, NULL, NETCF_EOTHER,
			"Failed to write interface '%s' xml representation to memory stream",
			(ifp && ifp->name ? ifp->name : NULL));
		goto error;
	}

	fclose(mem_stream);
	xml_node_free(xml);

	return out_string;

error:
	if(mem_stream)
		fclose(mem_stream);
	if(xml)
		xml_node_free(xml);
	return NULL;
}

static nc_interface_t *
__ncf_if_state(struct netcf_if *nif)
{
	struct netcf *ncf = nif->netcf;
	nc_interface_t *ifp;

	if(__ncf_refresh(&ncf->system) < 0) {
		__ncf_error_report(ncf, &ncf->system, NETCF_EOTHER,
			"Cannot read system interface state");
		return NULL;
	}
	ifp = nc_interface_by_name(ncf->system.handle, nif->ifp->name);
	if(!ifp) {
		/* not active (any more) */
		__ncf_error_clear(ncf, &ncf->system);
		return NULL;
	}
	return ifp;
}

static int
__ncf_refresh(struct netcf_ifinfo *nii)
{
	/* some sort of (mtime) caching check?! */
	if(nc_interfaces_refresh(nii->handle) < 0) {
		nc_error("interface refresh failed");
		return -1;
	}
	return 0;
}

